########### R function: threeLevelMargLogLik ###########

# For the three level lower bound of the marginal log-likelihood.

# Last changed: 18 JUN 2020

threeLevelMargLogLik <- function(p, m, q1, q2, nVec, oMat, XList, L1List, L2List,
                                 two_norm_terms, uL1_tcrossprod_terms, uL2_tcrossprod_terms,
                                 nu.sigsq, s.sigsq, mu.q.recip.sigsq, xi.q.sigsq, lambda.q.sigsq,
                                 mu.q.recip.a.sigsq, xi.q.a.sigsq, lambda.q.a.sigsq, Sigma.beta,
                                 mu.beta, mu.q.beta, Sigma.q.beta,  M.q.inv.SigmaL1, xi.q.SigmaL1,
                                 Lambda.q.SigmaL1, nu.SigmaL1, s.SigmaL1, M.q.inv.A.SigmaL1,
                                 xi.q.A.SigmaL1, Lambda.q.A.SigmaL1, M.q.inv.SigmaL2, xi.q.SigmaL2,
                                 Lambda.q.SigmaL2, nu.SigmaL2, s.SigmaL2, M.q.inv.A.SigmaL2,
                                 xi.q.A.SigmaL2, Lambda.q.A.SigmaL2)
{

    # Compute other terms of the marginal log-likelihood:

    T1 <-  ( - 0.5 * log(2 * pi) * sum(oMat)
             - 0.5 * mu.q.recip.sigsq * two_norm_terms )

    T2 <- ( - 0.5 * (p + (m * q1) + (sum(nVec)*q2)) * log(2 * pi)
            - 0.5 * as.numeric( determinant( Sigma.beta )$modulus)
            - 0.5 * sum(diag((tcrossprod(mu.q.beta - mu.beta) + Sigma.q.beta) %*% solve(Sigma.beta)))
            - 0.5 * sum(diag(M.q.inv.SigmaL1 %*% uL1_tcrossprod_terms))
            - 0.5 * sum(diag(M.q.inv.SigmaL2 %*% uL2_tcrossprod_terms)))

    # Calculate streamlined version of log det Sigma:

    termA <- as.numeric(determinant(Sigma.q.beta)$modulus)

    termB <- 0
    for (i in 1:m)
    {
       ZL1CrossProdTerm <- 0
       for (j in 1:nVec[i])
          ZL1CrossProdTerm <- ZL1CrossProdTerm + crossprod(ZL1List[[i]][[j]])

       term1 <- mu.q.recip.sigsq*ZL1CrossProdTerm + M.q.inv.SigmaL1

       term2 <- 0 ; term3 <- 0
       for (j in 1:nVec[i])
       {

          crossprodL2Term <- mu.q.recip.sigsq*crossprod(ZL2List[[i]][[j]]) + M.q.inv.SigmaL2

          term2 <- term2 + ((mu.q.recip.sigsq^2)*crossprod(XList[[i]][[j]],
                            ZL2List[[i]][[j]]) %*% solve(crossprodL2Term) %*%
                            crossprod(ZL2List[[i]][[j]], XList[[i]][[j]]))
          term3 <- term3 + as.numeric(determinant(crossprodL2Term)$modulus)
       }
       termB <- (termB
                 + as.numeric(determinant(term1 - term2)$modulus)
                 + term3)
    }
    log.det.Sigma.q.betau <- termA - termB

    T3 <- ( 0.5 * (p + (m * q1) + (sum(nVec)*q2))
            + 0.5 * (p + (m * q1) + (sum(nVec)*q2)) * log(2 * pi)
            + 0.5 * log.det.Sigma.q.betau)

    T4 <- ( - 0.5 * nu.sigsq * log(2)
            - lgamma(0.5 * nu.sigsq)
            - 0.5 * mu.q.recip.a.sigsq * mu.q.recip.sigsq )

    T5 <- ( - 0.5 * xi.q.sigsq * log(0.5 * lambda.q.sigsq)
            + lgamma(0.5 * xi.q.sigsq)
            + 0.5 * lambda.q.sigsq * mu.q.recip.sigsq )

    T6 <- ( - 0.5 * log(2 * nu.sigsq * (s.sigsq^2))
            - lgamma(0.5)
            - (1/(2 * nu.sigsq * (s.sigsq^2))) * mu.q.recip.a.sigsq )

    T7 <- ( - 0.5 * xi.q.a.sigsq * log(0.5 * lambda.q.a.sigsq)
            + lgamma(0.5 * xi.q.a.sigsq)
            + 0.5 * lambda.q.a.sigsq * mu.q.recip.a.sigsq )

    T8 <- ( - 0.5 * sum( diag( M.q.inv.A.SigmaL1 %*% M.q.inv.SigmaL1 ) )
            - 0.5 * q1 * (nu.SigmaL1 + q1 - 1) * log(2)
            - (q1/4) * (q1 - 1) * log(pi)
            - sum( lgamma( 0.5 * ( nu.SigmaL1 + (2 * q1) - (1:q1) ) ) ) )

    T9 <- ( - 0.5 * ( xi.q.SigmaL1 - q1 + 1 ) * as.numeric( determinant( Lambda.q.SigmaL1 )$modulus )
            + 0.5 * sum( diag( Lambda.q.SigmaL1 %*% M.q.inv.SigmaL1 ) )
            + 0.5 * q1 * (xi.q.SigmaL1 - q1 + 1) * log(2)
            + (q1/4) * (q1 - 1) * log(pi)
            + sum( lgamma( 0.5 * ( xi.q.SigmaL1 + 2 - q1 - (1:q1) ) ) ) )

    T10 <- (- 0.5 * (1/(nu.SigmaL1)) * sum((1/(s.SigmaL1^2)) * diag(M.q.inv.A.SigmaL1))
                - 0.5 * q1 * log(nu.SigmaL1)
                - 0.5 * sum(log(s.SigmaL1^2))
                - 0.5 * q1 * log(2)
                - q1 * lgamma(0.5))

    T11 <- ( - 0.5 * xi.q.A.SigmaL2 * as.numeric( determinant( Lambda.q.A.SigmaL2 )$modulus )
             + 0.5 * sum( diag( Lambda.q.A.SigmaL2 %*% M.q.inv.A.SigmaL2 ) )
             + 0.5 * q1 * xi.q.A.SigmaL2 * log(2)
             + q1 * lgamma(0.5 * xi.q.A.SigmaL2))

    T12 <- ( - 0.5 * sum( diag( M.q.inv.A.SigmaL2 %*% M.q.inv.SigmaL2 ) )
            - 0.5 * q2 * (nu.SigmaL2 + q2 - 1) * log(2)
            - (q2/4) * (q2 - 1) * log(pi)
            - sum( lgamma( 0.5 * ( nu.SigmaL2 + (2 * q2) - (1:q2) ) ) ) )

    T13 <- ( - 0.5 * ( xi.q.SigmaL2 - q2 + 1 ) * as.numeric( determinant( Lambda.q.SigmaL2 )$modulus )
            + 0.5 * sum( diag( Lambda.q.SigmaL2 %*% M.q.inv.SigmaL2 ) )
            + 0.5 * q2 * (xi.q.SigmaL2 - q2 + 1) * log(2)
            + (q2/4) * (q2 - 1) * log(pi)
            + sum( lgamma( 0.5 * ( xi.q.SigmaL2 + 2 - q2 - (1:q2) ) ) ) )

    T14 <- (- 0.5 * (1/(nu.SigmaL2)) * sum((1/(s.SigmaL2^2)) * diag(M.q.inv.A.SigmaL2))
            - 0.5 * q2 * log(nu.SigmaL2)
            - 0.5 * sum(log(s.SigmaL2^2))
            - 0.5 * q2 * log(2)
            - q2 * lgamma(0.5))

    T15 <- ( - 0.5 * xi.q.A.SigmaL2 * as.numeric( determinant( Lambda.q.A.SigmaL2 )$modulus )
            + 0.5 * sum( diag( Lambda.q.A.SigmaL2 %*% M.q.inv.A.SigmaL2 ) )
            + 0.5 * q2 * xi.q.A.SigmaL2 * log(2)
            + q2 * lgamma(0.5 * xi.q.A.SigmaL2))

    logMlVal <- ( T1 + T2 + T3 + T4 + T5 + T6 + T7 + T8 + T9 + T10 + T11
                   + T12 + T13 + T14 + T15 )

    return(logMlVal)
}

########### End of threeLevelMargLogLik ############
