########### R function: twoLevelMargLogLik ###########

# For the lower bound of the marginal log-likelihood.

# Last changed: 15 JUN 2020

twoLevelMargLogLik <- function(p, m, q, nVec, ZList, two_norm_terms, u_tcrossprod_terms,
                                nu.sigsq, s.sigsq, mu.q.recip.sigsq, xi.q.sigsq, lambda.q.sigsq,
                                mu.q.recip.a.sigsq, xi.q.a.sigsq, lambda.q.a.sigsq,
                                Sigma.beta, mu.beta, mu.q.beta, Sigma.q.beta,
                                M.q.inv.Sigma, xi.q.Sigma, Lambda.q.Sigma, nu.Sigma, s.Sigma,
                                M.q.inv.A.Sigma, xi.q.A.Sigma, Lambda.q.A.Sigma)
{
    T1 <-  ( - 0.5 * log(2 * pi) * sum(nVec)
             - 0.5 * mu.q.recip.sigsq * two_norm_terms )

    T2 <- ( - 0.5 * (p + (m * q)) * log(2 * pi)
            - 0.5 * as.numeric( determinant( Sigma.beta )$modulus)
            - 0.5 * sum(diag((tcrossprod(mu.q.beta - mu.beta) + Sigma.q.beta) %*% solve(Sigma.beta)))
            - 0.5 * sum(diag(M.q.inv.Sigma %*% u_tcrossprod_terms)) )

    termA <- as.numeric(determinant(Sigma.q.beta)$modulus)
    termB <- 0
    for (i in 1:m)
    {
       logDetTerm <- mu.q.recip.sigsq * crossprod(ZList[[i]]) + M.q.inv.Sigma
       termB <- termB + as.numeric(determinant(logDetTerm)$modulus)
    }
    log.det.Sigma.q.betau <- termA - termB

    T3 <- ( 0.5 * (p + (m * q))
            + 0.5 * (p + (m * q)) * log(2 * pi)
            + 0.5 * log.det.Sigma.q.betau)

    T4 <- ( - 0.5 * nu.sigsq * log(2)
            - lgamma(0.5 * nu.sigsq)
            - 0.5 * mu.q.recip.a.sigsq * mu.q.recip.sigsq )

    T5 <- ( - 0.5 * xi.q.sigsq * log(0.5 * lambda.q.sigsq)
            + lgamma(0.5 * xi.q.sigsq)
            + 0.5 * lambda.q.sigsq * mu.q.recip.sigsq )

    T6 <- ( - 0.5 * log(2 * nu.sigsq * (s.sigsq^2))
            - lgamma(0.5)
            - (1/(2 * nu.sigsq * (s.sigsq^2))) * mu.q.recip.a.sigsq )

    T7 <- ( - 0.5 * xi.q.a.sigsq * log(0.5 * lambda.q.a.sigsq)
            + lgamma(0.5 * xi.q.a.sigsq)
            + 0.5 * lambda.q.a.sigsq * mu.q.recip.a.sigsq )

    T8 <- ( - 0.5 * sum( diag( M.q.inv.A.Sigma %*% M.q.inv.Sigma ) )
            - 0.5 * q * (nu.Sigma + q - 1) * log(2)
            - (q/4) * (q - 1) * log(pi)
            - sum( lgamma( 0.5 * ( nu.Sigma + (2 * q) - (1:q) ) ) ) )

    T9 <- ( - 0.5 * ( xi.q.Sigma - q + 1 ) * as.numeric( determinant( Lambda.q.Sigma )$modulus )
            + 0.5 * sum( diag( Lambda.q.Sigma %*% M.q.inv.Sigma ) )
            + 0.5 * q * (xi.q.Sigma - q + 1) * log(2)
            + (q/4) * (q - 1) * log(pi)
            + sum( lgamma( 0.5 * ( xi.q.Sigma + 2 - q - (1:q) ) ) ) )

    T10 <- (- 0.5 * (1/(nu.Sigma)) * sum((1/(s.Sigma^2)) * diag(M.q.inv.A.Sigma))
                - 0.5 * q * log(nu.Sigma)
                - 0.5 * sum(log(s.Sigma^2))
                - 0.5 * q * log(2)
                - q * lgamma(0.5))

    T11 <- ( - 0.5 * xi.q.A.Sigma * as.numeric( determinant( Lambda.q.A.Sigma )$modulus )
             + 0.5 * sum( diag( Lambda.q.A.Sigma %*% M.q.inv.A.Sigma ) )
             + 0.5 * q * xi.q.A.Sigma * log(2)
             + q * lgamma(0.5 * xi.q.A.Sigma))

    logMlVal <- ( T1 + T2 + T3 + T4 + T5 + T6 + T7 + T8 + T9 + T10 + T11 )

    return(logMlVal)
}


########### End of twoLevelMargLogLik ############
